#!/user/bin/env python
'''containsDProteinChain.py

This filter returns entries that contain protein chain(s) made of D-amino acids.
The default constructor returns entries that contain at least one
polymer chain that is an D-protein. If the "exclusive" flag is set to true
in the constructor, all polymer chains must be D-proteins. For a multi-model structure,
this filter only checks the first model.

'''

__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

from mmtfPyspark.filters import ContainsPolymerChainType


class ContainsDProteinChain(object):
    '''Default constructor matches any entry that contains at least one D-protein chain.
    As an example a D-protein/DNA complex passes this filter

    Optional constructor that can be used to filter entries that exclusively
    contain D-protein chains. For example, with "exclusive" set to true, a
    D-protein/DNA complex does not pass this filter.

    Attributes
    ----------
    exclusive : bool
       if true, only return entries that are exclusively contain D-protein chains
    '''

    def __init__(self, exclusive=False):
        self.filter = ContainsPolymerChainType([
            ContainsPolymerChainType.D_PEPTIDE_LINKING,
            ContainsPolymerChainType.PEPTIDE_LINKING], exclusive=exclusive)

    def __call__(self, t):
        return self.filter(t)



#!/user/bin/env python
__author__ = "Peter Rose"
__version__ = "0.2.0"

from enum import Enum

from pyspark import SparkFiles
from pyspark.sql import SparkSession
from pyspark.sql.functions import split


class PtmType(Enum):
    ACETYLATION = 'Acetylation'
    PHOSPHORYLATON = 'Phosphorylation'
    UBIQUITINATION = 'Ubiquitination'
    SUCCINYLATION = 'Succinylation'
    METHYLATION = 'Methylation'
    MALONYLATION = 'Malonylation'
    N_LINKEDGLYCOSYLATION = 'N-linkedGlycosylation'
    O_LINKEDGLYCOSYLATION = 'O-linkedGlycosylation'
    SUMOYLATION = 'Sumoylation'
    S_NITROSYLATION = 'S-nitrosylation'
    GLUTATHIONYLATION = 'Glutathionylation'
    AMIDATION = 'Amidation'
    PYRROLIDONECARBOXYLICACID = 'PyrrolidoneCarboxylicAcid'
    PALMITOYLATION = 'Palmitoylation'
    GAMMA_CARBOXYGLUTAMICACID = 'Gamma-carboxyglutamicAcid'
    CROTONYLATION = 'Crotonylation'
    OXIDATION = 'Oxidation'
    MYRISTOYLATION = 'Myristoylation'
    C_LINKEDGLYCOSYLATION = 'C-linkedGlycosylation'
    SULFATION = 'Sulfation'
    FORMYLATION = 'Formylation'
    CITRULLINATION = 'Citrullination'
    GPI_ANCHOR = 'GPI-anchor'
    NITRATION = 'Nitration'
    S_DIACYLGLYCEROL = 'S-diacylglycerol'
    LIPOYLATION = 'Lipoylation'
    CARBAMIDATION = 'Carbamidation'
    NEDDYLATION = 'Neddylation'
    PYRUVATE = 'Pyruvate'
    S_LINKEDGLYCOSYLATION = 'S-linkedGlycosylation'


# location of dbPTM data
DB_PTM_URL = "http://dbptm.mbc.nctu.edu.tw/download/experiment/"
# location of cached dataset (temporary location for prototyping)
CACHED_FILE_URL = "https://github.com/sbl-sdsc/mmtf-data/raw/master/data/dbptm.orc.lzo"
FILENAME = "dbptm.orc.lzo"


def get_ptm_dataset():
    """
    Returns a cached dataset of all post-translational modifications in dbPTM. The dataset
    contains the following columns:
    +-----------+---------+-------------+---------------+--------------------+--------------------+
    |uniProtName|uniProtId|uniProtSeqNum|        ptmType|           pubMedIds|     sequenceSegment|
    +-----------+---------+-------------+---------------+--------------------+--------------------+
    |14310_ARATH|   P48347|          209|Phosphorylation|[23328941, 23572148]|AFDDAIAELDSLNEESY...|
    |14310_ARATH|   P48347|          233|Phosphorylation|          [23572148]|QLLRDNLTLWTSDLNEE...|
    |14310_ARATH|   P48347|          234|Phosphorylation|          [18463617]|LLRDNLTLWTSDLNEEG...|

    Reference:
    dbPTM: http://dbptm.mbc.nctu.edu.tw/

    Kai-Yao Huang, Min-Gang Su, Hui-Ju Kao, Yun-Chung Hsieh, Jhih-Hua Jhong, Kuang-Hao Cheng, Hsien-Da Huang and
    Tzong-Yi Lee. "dbPTM 2016: 10-year anniversary of a resource for post-translational modification of proteins"
    Nucleic Acids Research, Volume 44, Issue D1, 4 January 2016, Pages D435-D446 - [PubMed]

    :return: dataset of PTMs in dbPTM
    """
    spark = SparkSession.builder.getOrCreate()

    # download cached dataset
    spark.sparkContext.addFile(CACHED_FILE_URL)

    # read dataset
    spark.conf.set("spark.sql.orc.impl", "native")
    return spark.read.orc(SparkFiles.get(FILENAME))


def download_ptm_dataset(ptm_type):
    """
    Downloads the specified type of post-translational modifications in dbPTM. The dataset
    contains the following columns:
    +-----------+---------+-------------+---------------+--------------------+--------------------+
    |uniProtName|uniProtId|uniProtSeqNum|        ptmType|           pubMedIds|     sequenceSegment|
    +-----------+---------+-------------+---------------+--------------------+--------------------+
    |14310_ARATH|   P48347|          209|Phosphorylation|[23328941, 23572148]|AFDDAIAELDSLNEESY...|
    |14310_ARATH|   P48347|          233|Phosphorylation|          [23572148]|QLLRDNLTLWTSDLNEE...|
    |14310_ARATH|   P48347|          234|Phosphorylation|          [18463617]|LLRDNLTLWTSDLNEEG...|

    Reference:
    dbPTM: http://dbptm.mbc.nctu.edu.tw/

    Kai-Yao Huang, Min-Gang Su, Hui-Ju Kao, Yun-Chung Hsieh, Jhih-Hua Jhong, Kuang-Hao Cheng, Hsien-Da Huang and
    Tzong-Yi Lee. "dbPTM 2016: 10-year anniversary of a resource for post-translational modification of proteins"
    Nucleic Acids Research, Volume 44, Issue D1, 4 January 2016, Pages D435-D446 - [PubMed]

    :return: dataset of PTMs in dbPTM
    """

    file_name = ptm_type.value + ".txt.gz"
    url = DB_PTM_URL + file_name
    print("downloading: " + url)

    spark = SparkSession.builder.getOrCreate()

    # download cached dataset
    spark.sparkContext.addFile(url)

    # read dataset
    ds = spark.read.option('header', 'true') \
        .option('sep', '\t') \
        .option('comment', '#') \
        .option('inferSchema', 'true') \
        .csv(SparkFiles.get(file_name))

    ds = ds.toDF("uniProtName", "uniProtId", "uniProtSeqNum", "ptmType", "pubMedIds", "sequenceSegment")
    ds = ds.withColumn("pubMedIds", split(ds.pubMedIds, ';'))

    return ds


def download_all_ptm_dataset():
    """
    Downloads all post-translational modifications in dbPTM. The dataset
    contains the following columns:
    +-----------+---------+-------------+---------------+--------------------+--------------------+
    |uniProtName|uniProtId|uniProtSeqNum|        ptmType|           pubMedIds|     sequenceSegment|
    +-----------+---------+-------------+---------------+--------------------+--------------------+
    |14310_ARATH|   P48347|          209|Phosphorylation|[23328941, 23572148]|AFDDAIAELDSLNEESY...|
    |14310_ARATH|   P48347|          233|Phosphorylation|          [23572148]|QLLRDNLTLWTSDLNEE...|
    |14310_ARATH|   P48347|          234|Phosphorylation|          [18463617]|LLRDNLTLWTSDLNEEG...|

    Reference:
    dbPTM: http://dbptm.mbc.nctu.edu.tw/

    Kai-Yao Huang, Min-Gang Su, Hui-Ju Kao, Yun-Chung Hsieh, Jhih-Hua Jhong, Kuang-Hao Cheng, Hsien-Da Huang and
    Tzong-Yi Lee. "dbPTM 2016: 10-year anniversary of a resource for post-translational modification of proteins"
    Nucleic Acids Research, Volume 44, Issue D1, 4 January 2016, Pages D435-D446 - [PubMed]

    :return: dataset of PTMs in dbPTM
    """
    ds = download_ptm_dataset(PtmType.PHOSPHORYLATON)

    for ptm in PtmType:
        if ptm != PtmType.PHOSPHORYLATON:
            ds = ds.union(download_ptm_dataset(ptm))

    return ds


def save_all_ptm_dataset():
    spark = SparkSession.builder.getOrCreate()
    spark.conf.set('spark.sql.orc.impl', 'native')
    ds = download_all_ptm_dataset()

    ds.coalesce(1).write.mode('overwrite').option('compression', 'lzo').orc("/Users/peter/work/dbPTM/dbptm.orc.lzo")



from .proteinSequenceEncoder import ProteinSequenceEncoder
from .sparkMultiClassClassifier import SparkMultiClassClassifier
from .sparkRegressor import SparkRegressor
from .sparkMultiClassClassifier import SparkMultiClassClassifier
from . import datasetBalancer, datasetClassifier, datasetRegressor, pythonRDDToDataset, sequenceNgrammer



from . import advancedSearchDataset, customReportService, dataset_utils, dbPtmDataset, dbSnpDataset, drugBankDataset, g2sDataset, jpredDataset, myVariantDataset, \
    pdbjMineDataset, pdbPtmDataset, pdbToUniProt, polymerSequenceExtractor, secondaryStructureElementExtractor, \
    secondaryStructureExtractor, secondaryStructureSegmentExtractor, swissModelDataset, uniProt
from .groupInteractionExtractor import groupInteractionExtractor



#!/user/bin/env python
"""
Downloads PDB to UniProt chain and residue-level mappings from the SIFTS project.
Data are retrieved from a data cache and only data that are not cached
are downloaded when required.
This class also provides a method to create a new mapping file.

For more information about SIFTS see:
The "Structure Integration with Function, Taxonomy and Sequence"
https://www.ebi.ac.uk/pdbe/docs/sifts/overview.html is
the authoritative source of up-to-date chain and residue-level
mappings to UniProt.

Example of chain-level mappings

>>> df = get_chain_mappings()
>>> df.show()

+----------------+-----------+-------+---------+
|structureChainId|structureId|chainId|uniprotId|
+----------------+-----------+-------+---------+
|          1A02.F|       1A02|      F|   P01100|
|          1A02.J|       1A02|      J|   P05412|
|          1A02.N|       1A02|      N|   Q13469|

Example of residue-level mappings

>>> df = get_cached_residue_mappings()
>>> df.show()

Columns:
structureChainId - pdbId.chainId
pdbResNum - PDB residue number in ATOM records
pdbSeqNum - PDB residue number in the sequence record (index start at 1)
uniprotId - UniProt id (accession number)
uniprotNum - UniProt residue number (index starts at 1)

+----------------+---------+---------+---------+----------+
|structureChainId|pdbResNum|pdbSeqNum|uniprotId|uniprotNum|
+----------------+---------+---------+---------+----------+
|          1STP.A|     null|        1|   P22629|        25|
|          1STP.A|     null|        2|   P22629|        26|
|          1STP.A|     null|        3|   P22629|        27|
|          1STP.A|     null|       12|   P22629|        36|
 ...
|          1STP.A|       13|       13|   P22629|        37|
|          1STP.A|       14|       14|   P22629|        38|
|          1STP.A|       15|       15|   P22629|        39|
 ...
"""

__author__ = "Peter Rose"
__version__ = "0.2.0"

from pyspark import SparkFiles
from pyspark.sql import SparkSession
from pyspark.sql.functions import concat_ws, upper

# location of cached dataset (temporary location for prototyping)
CACHED_FILE_URL = "https://github.com/sbl-sdsc/mmtf-data/raw/master/data/pdb2uniprot_residues.orc.lzo"
FILENAME = "pdb2uniprot_residues.orc.lzo"

# location of SIFTS data
UNIPROT_MAPPING_URL = "http://ftp.ebi.ac.uk/pub/databases/msd/sifts/flatfiles/csv/pdb_chain_uniprot.csv.gz"
UNIPROT_FILE = "pdb_chain_uniprot.csv.gz"
SIFTS_URL = "http://ftp.ebi.ac.uk/pub/databases/msd/sifts/split_xml/"


def get_chain_mappings(ids=None):
    """
    Returns an up-to-date dataset of PDB to UniProt
    chain-level mappings for a list of ids.
    Valid ids are either a list of pdbIds (e.g. 1XYZ) or pdbId.chainIds (e.g., 1XYZ.A).

    :param ids: list of pdbIds or pdbId.chainIds (pdbIds must be upper case, chainIds are case sensitive)
    :return: dataset of PDB to UniProt chain-level mappings
    """
    spark = SparkSession.builder.getOrCreate()

    # get a dataset of up-to-date UniProt chain mappings
    ds = __get_all_mappings()

    # return all mappings if no ids where passed in
    if not ids:
        return ds

    # convert list of ids to a list of lists (required for dataframe creation below)
    id_list = [[i] for i in ids]

    # create a dataset of ids from the passed-in list
    subset = spark.createDataFrame(id_list, ['id'])

    # create a subset of data for the list of ids
    if len(ids[0]) == 4:
        # join by pdbId
        ds = ds.join(subset, ds.pdbId == subset.id).drop("id")

    else:
        # join by pdbChainId
        ds = ds.join(subset, ds.pdbChainId == subset.id).drop("id")

    return ds


def get_cached_residue_mappings():
    """
    Returns the current version of the cached dataset of PDB to UniProt
    residue mappings. This method is faste, but may not
    contain the mapping for recently released PDB entries.
    :return: dataset of PDB to UniProt residue mappings
    """
    spark = SparkSession.builder.getOrCreate()
    spark.conf.set("spark.sql.orc.impl", "native")

    # download cached dataset
    spark.sparkContext.addFile(CACHED_FILE_URL)

    # read dataset
    ds = spark.read.format("orc").load(SparkFiles.get(FILENAME))

    return ds


def __get_all_mappings():
    spark = SparkSession.builder.getOrCreate()

    # download data
    spark.sparkContext.addFile(UNIPROT_MAPPING_URL)

    # read dataset
    ds = spark.read.option("header", "true") \
        .option("comment", "#") \
        .option("inferSchema", "true") \
        .csv(SparkFiles.get(UNIPROT_FILE))

    # cleanup and rename columns to be consistent with MMTF conventions.
    ds = ds.withColumn("PDB", upper(ds.PDB)) \
        .withColumnRenamed("PDB", "pdbId") \
        .withColumnRenamed("CHAIN", "chainId") \
        .withColumnRenamed("SP_PRIMARY", "uniProtId") \
        .withColumn("pdbChainId", concat_ws(".", "pdbId", "chainId")) \
        .select("pdbChainId", "pdbId", "chainId", "uniProtId")

    return ds



#!/user/bin/env python
'''orFilter.py

This filter wraps two filter and returns true if one of the filters passes

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "done"


class OrFilter(object):
    '''Constructor takes another filter as input

    Attributes
    ----------
    filter1 : filter
       first filter to be negated
    filter2 : filter
       second filter to be negated
    '''

    def __init__(self, filter1, filter2):
        self.filter1 = filter1
        self.filter2 = filter2

    def __call__(self, t):
        return self.filter1(t) or self.filter2(t)



#!/user/bin/env python
"""
advancedSearchDataset.py:

Runs an RCSB PDB Advanced Search web service using an XML query description.
See https://www.rcsb.org/pdb/staticHelp.do?p=help/advancedSearch.html Advanced Search for
an overview and a list of available queries at
https://www.rcsb.org/pdb/staticHelp.do?p=help/advancedsearch/index.html
The returned dataset contains the following field dependent on the query type:

# structureId, e.g., 4HHB
# structureChainId, e.g., 4HHB.A
# ligandId, e.g., HEM
"""
__author__ = "Peter Rose"
__version__ = "0.2.0"

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import concat_ws, explode, split, substring_index

from mmtfPyspark.datasets import pdbjMineDataset
from mmtfPyspark.webservices.advancedQueryService import post_query


def get_dataset(xmlQuery):
    """
    Runs an RCSB PDB Advanced Search web service using an XML query description.
    See https://www.rcsb.org/pdb/staticHelp.do?p=help/advancedSearch.html Advanced Search
    The returned dataset contains the following field dependent on the query type:
    # structureId, e.g., 1STP
    # structureChainId, e.g., 4HHB.A
    # ligandId, e.g., HEM

    :param xmlQuery: RCSB PDB advanced query xml string
    :return: dataset with matching ids
    """

    # run advanced query
    ids = post_query(xmlQuery)

    # convert list of ids to a list of lists (required for dataframe creation below)
    id_list = [[i] for i in ids]

    # convert list of lists to a dataframe
    spark = SparkSession.builder.getOrCreate()

    # distinguish 3 types of results based on length of string
    # structureId: 4 (e.g., 4HHB)
    # structureEntityId: > 4 (e.g., 4HHB:1)
    # entityId: < 4 (e.g., HEM)

    if len(ids[0]) > 4:
        ds: DataFrame = spark.createDataFrame(id_list, ['pdbEntityId'])
        # if results contain an entity id, e.g., 101M:1, then map entityId to pdbChainId
        ds = ds.withColumn("pdbId", substring_index(ds.pdbEntityId, ':', 1))
        ds = ds.withColumn("entityId", substring_index(ds.pdbEntityId, ':', -1))
        mapping = __get_entity_to_chain_id()
        ds = ds.join(mapping, (ds.pdbId == mapping.structureId) & (ds.entityId == mapping.entity_id))
        ds = ds.select(ds.pdbChainId)
    elif len(ids[0]) < 4:
        ds: DataFrame = spark.createDataFrame(id_list, ['ligandId'])
    else:
        ds: DataFrame = spark.createDataFrame(id_list, ['pdbId'])

    return ds


def __get_entity_to_chain_id():
    # get entityID to strandId mapping
    query = "SELECT pdbid, entity_id, pdbx_strand_id FROM entity_poly"
    mapping: DataFrame = pdbjMineDataset.get_dataset(query)

    # split one-to-many relationship into multiple records: 'A,B -> [A, B] -> explode to separate rows
    mapping = mapping.withColumn("chainId", split(mapping.pdbx_strand_id, ","))
    mapping = mapping.withColumn("chainId", explode("chainId"))

    # create a structureChainId file, e.g. 1XYZ + A -> 1XYZ.A
    mapping = mapping.withColumn("pdbChainId", concat_ws(".", mapping.structureId, mapping.chainId))

    return mapping.select(mapping.entity_id, mapping.structureId, mapping.pdbChainId)



#!/user/bin/env python
"""customReportService.py

This class uses RCSB PDB Tabular Report RESTful webservices to retrieve
metadata and annotations for all current entries in the ProteinDataBank.

References
----------
- List of supported fieldnames: http://www.rcsb.org/pdb/results/reportField.do
- The RCSB Protein Data Bank:redesignedwebsiteandwebservices2011NucleicAcidsRes.39:D392-D401. https://dx.doi.org/10.1093/nar/gkq1021

Examples
--------
Retrieve PubMedCentral, PubMedID, and Depositiondate:

>>> ds = customReportService.get_dataset("pmc","pubmedId","depositionDate")
>>> ds.printSchema()
>>> ds.show(5)
"""

__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

import tempfile
import urllib
from urllib import request

from pyspark.sql import SparkSession

SERVICELOCATIONS = ["https://www.rcsb.org/pdb/rest/customReport",
                    "https://www1.rcsb.org/pdb/rest/customReport",
                    "https://www2.rcsb.org/pdb/rest/customReport"]


CURRENT_URL = "?pdbids=*&service=wsfile&format=csv&primaryOnly=1&customReportColumns="


def get_dataset(columnNames):
    """Returns a dataset with the specified columns for all current PDB entires.
    See <a href="https://www.rcsb.org/pdb/results/reportField.do">  for a list
    of supported filed names

    Parameters
    ----------
    columnNames : str, list
       names of columns for the dataset

    Returns
    -------
    dataset
       dataset with the specified columns
    """
    if type(columnNames) == str:
        columnNames = [columnNames]
    query = CURRENT_URL + ','.join(columnNames)
    return _get_dataset(query, columnNames)


def _get_dataset(query, columnNames):
    """Get dataset using different service locations"""

    dataset = None
    spark = None
    for SERVICELOCATION in SERVICELOCATIONS:
        try:
            inStream = _post_query(SERVICELOCATION, query)

            tmp = tempfile.NamedTemporaryFile(delete=False)

            with open(tmp.name, "w") as t:
                    for l in inStream:
                        t.writelines(str(l)[2:-3] + '\n')

            spark = SparkSession.builder.getOrCreate()
            dataset = _read_csv(spark, tmp.name)
        except:
            continue
        break

    if dataset is None:
        print("ERROR: cannot connect to service location")
        return dataset

    return _concat_ids(spark, dataset, columnNames)


def _concat_ids(spark, dataset, columnNames):
    """Concatenates structureId and chainId fields into a single key if chainId
    field is present

    Parameters
    ----------
    spark : :obj:`SparkSession <pyspark.sql.SparkSession>`
    dataset : Dataframe
    columnNames : list
    """

    if "chainId" in dataset.columns:
        dataset.createOrReplaceTempView("table")

        sql = "SELECT CONCAT(structureId,'.',chainId) as structureChainId," + \
              "structureId,chainId,%s" % ','.join(columnNames) + \
              " from table"

        dataset = spark.sql(sql)

    return dataset


def _post_query(service, query):
    """Post PDB Ids and fields in a query string to the RESTful RCSB web service

    Parameters
    ----------
    query : str
       RESTful query urlopen
    service :str 
       service location

    Returns
    -------
    stream
        input stream to response
    """

    encodedQuery = urllib.parse.quote(query).encode('utf-8')
    url = request.Request(service)
    stream = urllib.request.urlopen(url, data=encodedQuery)

    return stream


def _read_csv(spark, inputFileName):
    '''Reads CSV file into a Spark dataset

    Parameters
    ----------
    spark : Spark Context
    inputFileName : str
       directory path for the input file
    '''

    dataset = spark.read \
                   .format("csv") \
                   .option("header", "true") \
                   .option("inferSchema", "true") \
                   .load(inputFileName) \
                   .cache()

    return dataset



#!/user/bin/env python
__author__ = "Peter Rose"
__version__ = "0.2.0"

from pyspark import SparkFiles
from pyspark.sql import SparkSession

# location of cached dataset (temporary location for prototyping)
CACHED_FILE_URL = "https://github.com/sbl-sdsc/mmtf-data/raw/master/data/nsSNP3D_PDB_UNP_GRCh37_t9606_s95.lzo.orc"
FILENAME = "nsSNP3D_PDB_UNP_GRCh37_t9606_s95.lzo.orc"


def get_cached_dataset():
    """
    Returns a dataset of missense mutations in dbSNP mapped to UniProt and PDB residue positions.

    The dataset contains the following columns:
    +---+---------+---------+------------+---------+----------+----------+----------+---------+-------+-------+ ...
    |chr|      pos|   snp_id|  master_acc|master_gi|master_pos|master_res|master_var|   pdb_gi|pdb_res|pdb_pos|
    +---+---------+---------+------------+---------+----------+----------+----------+---------+-------+-------+ ...
    |  4| 79525461|764726341|   NP_005130|  4826643|       274|         R|         *|157829892|      R|    274|
    |  4| 79525462|771966889|   NP_005130|  4826643|       274|         R|         P|157829892|      R|    274| ...

    ... +-----------+--------------------+----------+------+---------+---------+----------+
        |blast_ident|             clinsig|pdbChainId|tax_id|pdbResNum|uniprotId|uniprotNum|
    ... +-----------+--------------------+----------+------+---------+---------+----------+
        |      100.0|                null|    1AII.A|  9606|      275|   P12429|       274|
    ... |      100.0|                null|    1AII.A|  9606|      275|   P12429|       274|

    Reference:
    dbSNP: https://www.ncbi.nlm.nih.gov/projects/SNP/


    :return: dataset of missense mutations in dbSNP
    """
    spark = SparkSession.builder.getOrCreate()

    # download cached dataset
    spark.sparkContext.addFile(CACHED_FILE_URL)

    # read dataset
    spark.conf.set("spark.sql.orc.impl", "native")
    return spark.read.orc(SparkFiles.get(FILENAME))



#!/user/bin/env python
'''polymerSequenceExtractor.py:

Creates a dataset of polymer sequences using the full sequence
used in the experiment (i.e., the "SEQRES" record in PDB files).

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Debug"

from mmtfPyspark.ml import pythonRDDToDataset
from mmtfPyspark.mappers import StructureToPolymerSequences
from pyspark.sql import Row


def get_dataset(structures):
    '''Returns a dataset of polymer sequence contained in PDB entries
    using the full sequence used in the experimnet
    (i.e., the "SEQRES" record in PDB files)

    Parameters
    ----------
    structures : pythonRDD
       a set of PDB structures

    Returns
    -------
    dataset
       dataset with interacting residue and atom information
    '''

    rows = structures.flatMap(StructureToPolymerSequences()) \
                     .map(lambda x: Row(x[0],x[1]))

    colNames = ["structureChainId", "sequence"]

    return pythonRDDToDataset.get_dataset(rows, colNames)



#!/user/bin/env python
'''interactionFilter.py

A filter to specify criteria for molecular interactions between a query and a
target within a macromolecular structure. The filter specifies criteria for the
query (e.g. a metal ion) and the target (e.g. amino acid redisudes).
Interaction criteria such as distance cutoff limit the nature of interactions to
be considered.

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "done"

import sys
import numpy as np


class InteractionFilter(object):
    ''' A filter to specify criteria for molecular interactions between a query
    and a target

    Attributes
    ----------
    distanceCutoff : float
       cutoff distance for filter [default: float max]
    normalizedbFactorCutoff : float
       normalized bFactor cutoff for filter [default: float max]
    minInteractions : int
       minimum number of interactions [default: 1]
    maxInteractions : int
       maximum number of interactions [default: 10]
    '''

    def __init__(self, distanceCutoff=sys.float_info.max -1,
                 normalizedbFactorCutoff=sys.float_info.max -1,
                 minInteractions=1, maxInteractions=10):

        # Set interaction criteria
        self._distanceCutoff = distanceCutoff
        self._normalizedbFactorCutoff = normalizedbFactorCutoff
        self._minInteractions = minInteractions
        self._maxInteractions = maxInteractions

        # Query criteria
        self._queryGroups = None
        self._queryElements = None
        self._queryAtomNames = None

        # Target criteria
        self._targetGroups = None
        self._targetElements = None
        self._targetAtomNames = None
        self._prohibitedTargetGroups = None

    def get_distance_cutoff(self):
        '''Get the maximum interaction distance. At least one pair of query and
        target atoms must be wintin in distance cutoff

        Returns
        -------
        int
           maximum interaction distance
        '''
        return self._distanceCutoff

    def set_distance_cutoff(self, distanceCutoff):
        '''Set the distnace cutoff for interacting atoms

        Parameters 
        ----------
        distanceCutoff : float
           the maximum distance for interacting atoms
        '''

        self._distanceCutoff = distanceCutoff

    def get_normalized_b_factor_cutoff(self):
        '''Gets the maximum normalized b-factor (z-score) cutoff for an atom and
        its interacting neighbor atoms.

        Returns
        -------
        float
           maximum normalized b-factor cutoff
        '''

        return self._normalizedbFactorCutoff

    def set_normalized_b_factor_cutoff(self, normalizedbFactorCutoff):
        '''Sets the maximum noramlized b-factor cutoff. This value represents a
        z-score (see reference below), the signed number of standard deviations
        by which the b-factor. High z-scores indicate either high flexibility
        and/or experimental error for the atoms involved in the interactions. By
        setting a cutoff value, not well defined interactions can be exculded.

        Frequently used z-scores:

        Confidence level    Tail area   z-scores
                     90%        0.05     +-1.645
                     95%        0.025    +-1.96
                     99%        0.005    +-2.576

        For example, to include all interactions within the 90% confidence
        interval, set the normalized b-factor to +1.645

        Parameters 
        ----------
        normalizedbFactorCutoff : float
           maximum normalized b-factor

        References
        ----------
        Z-score: https://en.wikipedia.org/wiki/Standard_score
        '''

        self._normalizedbFactorCutoff = normalizedbFactorCutoff

    def get_min_interactions(self):
        '''Returns the minimum number of interactions per atom. Atoms that interact
        with fewer atoms wil be discarded.

        Returns
        -------
        int
           minimum number of interactions per atom
        '''

        return self._minInteractions

    def set_min_interactions(self, minInteractions):
        '''Sets the minimum number of interactions per atom. Atoms that interact
        with fewer atoms will be discarded

        Parameters 
        ----------
        minInteractions : int
           minimum number of interactions per atom
        '''

        self._minInteractions = minInteractions

    def get_max_interactions(self):
        '''Returns the maximum number of interactions per atom. Atoms that interact
        with fewer atoms wil be discarded.

        Returns
        -------
        int
           maximum number of interactions per atom
        '''

        return self._maxInteractions

    def set_max_interactions(self, maxInteractions):
        '''Sets the maximum number of interactions per atom. Atoms that interact
        with fewer atoms will be discarded

        Parameters 
        ----------
        maxInteractions : int
           minimum number of interactions per atom
        '''

        self._maxInteractions = maxInteractions

    def set_query_elements(self, include, elements):
        '''Sets the elements to either be included or excluded in the query.
        Element strings are case sensitive (e.g., "Zn" for Zinc).

        Examples
        --------
        Only use elements O, N, S in the query groups in find polar interactions

        >>> filter = InteractionFilter()
        >>> filter.set_query_elements(True, ["O", "N", "S"])

        Exclude non-polar elements and hydrogen in query groups and use
        all other elements to find interactions.

        >>> elements = ['C', 'H', 'P']
        >>> filter.set_query_elements(False, elements)

        Parameters 
        ----------
        include : bool
           if True, uses the specifed elements in the query,
           if False, ignores the specified elemetns and use all other elements
        elements : list
           list of elements to be included or excluded in query
        '''

        if self._queryElements is not None:
            raise ValueError("ERROR: QueryElements have already been set.")

        if type(elements) == str:
            elements = [elements]
        self._includeQueryElements = include
        self._queryElements = set(elements)

    def set_target_elements(self, include, elements):
        '''Sets the elements to either be included or excluded in the target.
        Element strings are case sensitive (e.g., "Zn" for Zinc).

        Examples
        --------
        Only use elements O, N, S in the target groups in find polar interactions

        >>> filter = InteractionFilter()
        >>> filter.set_target_elements(True, ["O", "N", "S"])

        Exclude non-polar elements and hydrogen in target groups and use
        all other elements to find interactions.

        >>> elements = ['C', 'H', 'P']
        >>> filter.set_query_elements(False, elements)

        Parameters 
        ----------
        include : bool
           if True, uses the specifed elements in the target,
           if False, ignores the specified elemetns and use all other elements
        elements : list
           list of elements to be included or excluded in target
        '''

        if self._targetElements is not None:
            raise ValueError("ERROR: TargetElements have already been set.")

        if type(elements) == str:
            elements = [elements]
        self._includeTargetElements = include
        self._targetElements = set(elements)

    def set_query_groups(self, include, groups):
        '''Sets groups to either be included or excluded in the query. Group names
        must be upper case (e.g. 'ZN' for Zinc)

        Examples
        --------
        Find interactions with ATP and ADP
        
        >>> filter = InteractionFilter()
        >>> filter.set_query_groups(True, ['ATP', 'ADP'])

        Exclude water and heavy water and use all other groups to find interactions.

        >>> groups = ["HOH", "DOD"]
        >>> filter.set_query_groups(False, groups)

        Parameters 
        ----------
        include : bool
           if True, uses the specified groups in the query,
           if False, ignores the specified groups and uses all other groups
        groups : list
           groups to be included or excluded in query

        '''

        if self._queryGroups is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(groups) == str:
            groups = [groups]
        self._includeQueryGroups = include
        self._queryGroups = set(groups)

    def set_target_groups(self, include, groups):
        '''Sets groups to either be included or excluded in the target. Group names
        must be upper case (e.g. 'ZN' for Zinc)

        Examples
        --------
        Find interactions with specific amino acid groups.

        >>> filter = InteractionFilter()
        >>> filter.set_target_groups(True, ['CYS','HIS','ASP','GLU'])

        Exclude water and heavy water and use all other groups to find interactions.

        >>> groups = ["HOH", "DOD"]
        >>> filter.set_target_groups(False, groups)

        Parameters 
        ----------
        include : bool
           if True, uses the specified groups in the target,
           if False, ignores the specified groups and uses all other groups
        groups : list
           groups to be included or excluded in query
        '''

        if self._targetGroups is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(groups) == str:
            groups = [groups]
        self._includeTargetGroups = include
        self._targetGroups = set(groups)

    def set_query_atom_names(self, include, atomNames):
        '''Set atom names to either be included or excluded in the query

        Examples
        --------
        Find interaction with C-alpha and C-beta atoms.
        
        >>> filter = InteractionFilter()
        >>> filter.set_query_atom_names(True, ['CA', 'CB'])

        Exclude backbone atoms, but consider all other atom names: e.g,
        amino acid side chains.

        >>> filter.set_query_atom_names(False, ['N', 'CA', 'C', 'O'])

        Parameters 
        ----------
        include : bool
           if True, uses set of atom names in query, 
           if False, ignores atoms with the specified names and uses all other atoms
        atom_names : list
           atoms to be included or excluded in the query
        '''

        if self._queryAtomNames is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(atomNames) == str:
            atomNames = [atomNames]
        self._includeQueryAtomNames = include
        self._queryAtomNames = set(atomNames)

    def set_target_atom_names(self, include, atomNames):
        '''Set atom names to either be included or excluded in the target.

        Examples
        --------
        Find interaction with C-alpha and C-beta atoms.

        >>> filter = InteractionFilter()
        >>> filter.set_target_atom_names(True, ['CA', 'CB'])

        Exclude backbone atoms, but consider all other atom names: e.g,
        amino acid side chains.

        >>> filter.set_target_atom_names(False, ['N', 'CA', 'C', 'O'])

        Parameters
        ----------
        include : bool
           if True, uses set of atom names in target, 
           if False, ignores atoms with the specified names and uses all other atoms
        atom_names : list
           atoms to be included or excluded in the target
        '''

        if self._targetAtomNames is not None:
            raise ValueError("ERROR: QueryGroups have already been set.")

        if type(atomNames) == str:
            atomNames = [atomNames]
        self._includeTargetAtomNames = include
        self._targetAtomNames = set(atomNames)

    def set_prohibited_target_groups(self, groups):
        '''Sets groups that must not appear in interactions. Any interactions that
        involves the specified groups will be excluded from the results.

        Examples
        --------
        Find Zinc interactions, but discard any interactions where the
        metal is involved in an interaction with water.

        >>> filter = InteractionFilter()
        >>> filter.set_query_groups(True, 'ZN')
        >>> filter.set_prohibited_target_groups(["HOH"])

        Parameters
        ----------
        groups : list
           one or more group names to be prohibited
        '''
        if type(groups) == str:
            groups = [groups]
        self._prohibitedTargetGroups = set(groups)

    def is_query_element(self, element):
        '''Returns True if the specified elements matches the query conditions.

        Parameters
        ----------
        element: str
           the element to be checked
        Returns
        -------
        bool
           True if element matches query conditinos, else False
        '''

        if self._queryElements is None:
            return True

        if self._includeQueryElements:
            return element in self._queryElements
        else:
            return element not in self._queryElements

    def is_target_element(self, element):
        '''Returns True if the specified elements matches the target conditions.

        Parameters
        ----------
        element : str
           the element to be checked

        Returns
        -------
        bool
           True if element matches target conditinos, else False
        '''

        if self._targetElements is None:
            return True

        if self._includeTargetElements:
            return element in self._targetElements
        else:
            return element not in self._targetElements

    def is_query_group(self, group):
        '''Returns True if the specified group matches the query conditions.

        Parameters
        ----------
        group : str
           the group to be checked

        Returns
        -------
        bool
           True if group matches query conditinos, else False

        '''

        if self._queryGroups is None:
            return True

        if self._includeQueryGroups:
            return group in self._queryGroups
        else:
            return group not in self._queryGroups

    def is_target_group(self, group):
        '''Returns True if the specified group matches the target conditions.

        Parameters
        ----------
        group : str
           the group to be checked

        Returns
        -------
        bool
           True if group matches target conditinos, else False
        '''

        if self._targetGroups is None:
            return True

        if self._includeTargetGroups:
            return group in self._targetGroups
        else:
            return group not in self._targetGroups

    def is_query_atom_name(self, atomName):
        '''Returns True if the specified atom matches the query conditions.

        Parameters
        ----------
        atomName : str
           the atom name to be checked

        Returns
        -------
        bool
           True if atom matches query conditinos, else False
        '''

        if self._queryAtomNames is None:
            return True

        if self._includeQueryAtomNames:
            return atomName in self._queryAtomNames
        else:
            return atomName not in self._queryAtomNames

    def is_target_atom_name(self, atomName):
        '''Returns True if the specified atom matches the target conditions.

        Parameters
        ----------
        atomName : str 
           the atom to be checked

        Returns
        -------
        bool
           True if atom matches target conditinos, else False
           
        '''

        if self._targetAtomNames is None:
            return True

        if self._includeTargetAtomNames:
            return atomName in self._targetAtomNames
        else:
            return atomName not in self._targetAtomNames

    def is_prohibited_target_group(self, group):
        '''Returns True if the specified group must not occur in an interactions.

        Parameters
        ----------
        group : str
           group that must not occur in interactions

        Returns
        -------
        bool
           True if group is prohibited else False
        '''

        if self._prohibitedTargetGroups is None:
            return False

        return group in self._prohibitedTargetGroups

    def is_query_element_np(self, elements):
        '''Returns True if the specified elements matches the query conditions.

        Parameters
        ----------
        element: str
           the element to be checked
        Returns
        -------
        bool
           True if element matches query conditinos, else False
        '''

        if self._queryElements is None:
            return np.full(elements.shape, True)
        else:
            inv = not self._includeQueryElements
            return np.in1d(elements, list(self._queryElements), invert=inv).reshape(elements.shape)

    def is_target_element_np(self, elements):
        '''Returns True if the specified elements matches the target conditions.

        Parameters
        ----------
        elements : str
           the element to be checked

        Returns
        -------
        bool
           True if element matches target conditions, else False
        '''

        if self._targetElements is None:
            return np.full(elements.shape, True)
        else:
            inv = not self._includeTargetElements
            return np.in1d(elements, list(self._targetElements), invert=inv).reshape(elements.shape)

    def is_query_group_np(self, groups):
        '''Returns True if the specified group matches the query conditions.

        Parameters
        ----------
        group : str
           the group to be checked

        Returns
        -------
        bool
           True if group matches query conditinos, else False

        '''
        if self._queryGroups is None:
            return np.full(groups.shape, True)
        else:
            inv = not self._includeQueryGroups
            return np.in1d(groups, list(self._queryGroups), invert=inv).reshape(groups.shape)

    def is_target_group_np(self, groups):
        '''Returns True if the specified group matches the target conditions.

        Parameters
        ----------
        group : str
           the group to be checked

        Returns
        -------
        bool
           True if group matches target conditinos, else False
        '''

        if self._targetGroups is None:
            return np.full(groups.shape, True)
        else:
            inv = not self._includeTargetGroups
            return np.in1d(groups, list(self._targetGroups), invert=inv).reshape(groups.shape)

    def is_query_atom_name_np(self, atomNames):
        '''Returns True if the specified atom matches the query conditions.

        Parameters
        ----------
        atomName : str
           the atom name to be checked

        Returns
        -------
        bool
           True if atom matches query conditinos, else False
        '''

        if self._queryAtomNames is None:
            return np.full(atomNames.shape, True)
        else:
            inv = not self._includeQueryAtomNames
            return np.in1d(atomNames, list(self._queryAtomNames), invert=inv).reshape(atomNames.shape)

    def is_target_atom_name_np(self, atomNames):
        '''Returns True if the specified atom matches the target conditions.

        Parameters
        ----------
        atomName : str
           the atom to be checked

        Returns
        -------
        bool
           True if atom matches target conditinos, else False

        '''

        if self._targetAtomNames is None:
            return np.full(atomNames.shape, True)
        else:
            inv = not self._includeTargetAtomNames
            return np.in1d(atomNames, list(self._targetAtomNames), invert=inv).reshape(atomNames.shape)


#!/user/bin/env python
'''rFree.py:

This filter returns true if the rFree value for this structure is within the
specified range

References
----------
- `rFree <http://pdb101.rcsb.org/learn/guide-to-understanding-pdb-data/r-value-and-r-free>`_

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"


class RFree(object):
    '''This filter returns True if the rFree value for this structure in withing
    the specified range.

    Attributes
    ----------
    min_Rfree : float
       The lower bound r_free value
    max_RFree : float
       The upper bound r_free value
    '''

    def __init__(self, minRfree, maxRfree):
        self.min_Rfree = minRfree
        self.max_Rfree = maxRfree

    def __call__(self, t):
        if t[1].r_free == None:
            return False

        return (t[1].r_free >= self.min_Rfree and t[1].r_free <= self.max_Rfree)



#!/user/bin/env python
'''blastCluster.py

This filter passes through representative structures from the RCSB PDB
BlastCLust cluster. A sequence identity thresholds needs to be specified.
The representative for each cluster is the first chain in a cluster.

References
----------
BlastClust cluster field names:
`Field names <http://www.rcsb.org/pdb/statistics/clusterStatistics.do>`_

Examples
--------
Find representative PDB entries at 90% sequence identity:
>>> sequenceIdentity = 90
>>> pdb = pdb.filter(BlastCluster(90))

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

import urllib.request


class BlastCluster(object):
	'''Filters blast clusters

	Attributes
	----------
	sequenceIdentity : int
	   sequence indentity for blast
	'''
	def __init__(self, sequenceIdentity):

		clusters = self.get_blast_cluster(sequenceIdentity)

		self.pdbIds = set()

		for protein in clusters:
			self.pdbIds.add(protein)
			self.pdbIds.add(protein[:4])


	def __call__(self, t):
		return t[0] in self.pdbIds


	def get_blast_cluster(self, sequenceIdentity):

		if sequenceIdentity not in [30,40,50,70,90,95,100]:
			raise Exception(f"Error: representative chains are not availible for \
							sequence Identity {sequenceIdentity}.\n Must be in \
							range [30,40,50,70,90,95,100]")
			return

		coreUrl = "https://cdn.rcsb.org/sequence/clusters/"
		clusters = []
		inputStream = urllib.request.urlopen(f"{coreUrl}bc-{sequenceIdentity}.out")

		for line in inputStream:
			line = str(line)[2:-3].replace("_",".").strip("\\n")
			clusters += line.split(" ")

		return clusters



#!/user/bin/env python
'''interaction_extractor.py

This class creates dataset of ligand - macromolecule and macromolecule -
macromolecule interaction information. Criteria to select interactions are
specified by the InteractionFilter.

'''
__author__ = "Peter W Rose"
__version__ = "0.3.0"
__status__ = "experimental"

from pyspark.sql import SparkSession
from pyspark.sql.types import *
from mmtfPyspark.utils import ColumnarStructure
from pyspark.sql import Row
import numpy as np
from scipy.spatial import cKDTree


class InteractionExtractor(object):

    @staticmethod
    def get_ligand_polymer_interactions(structures, interaction_filter, level='group'):
        '''Returns a dataset of ligand - macromolecule interactions

        The dataset contains the following columns:
        - structureChainId - pdbId.chainName of chain that interacts with ligand
        - queryLigandId - id of ligand from PDB chemical component dictionary
        - queryLigandNumber - group number of ligand including insetion code
        - queryLigandChainId - chain name of ligand
        - targetChainId - name of chain for which the interaction data are listed
        - groupNumbers - array of residue number of interacting groups including insertion code (e.g. 101A)
        - sequenceIndices - array of zero-based index of interaction groups (residues) mapped onto target sequence
        - sequence - interacting polymer sequence
        - interactingChains - total number of chains that interact with ligand

        Parameters
        ----------
        structures : PythonRDD
           a set of PDB structures
        interaction_filter : InteractionFilter
           interaction criteria
        level : 'group' or 'atom' to aggregate results

        Returns
        -------
        dataset
           dataset with interacting residue information
        '''

        # find sll interactions
        row = structures.flatMap(LigandInteractionFingerprint(interaction_filter, level))

        # Convert RDD to a Dataset with the following columns and types
        nullable = False

        # consider adding parameters
        # chem: add element, entity_type(LGO, PRO, DNA, etc.)
        # geom=True -> add distance, order parameters([q3,q4,q5,q6]
        # seq=True -> add sequence index, sequence

        fields = []
        if level == 'group':
            fields = [StructField("structureChainId", StringType(), nullable),
                      StructField("queryLigandId", StringType(), nullable),
                      StructField("queryLigandChainId", StringType(), nullable),
                      StructField("queryLigandNumber", StringType(), nullable),
                      StructField("targetGroupId", StringType(), nullable),
                      StructField("targetChainId", StringType(), nullable),
                      StructField("targetGroupNumber", StringType(), nullable),
                      StructField("sequenceIndex", IntegerType(), nullable),
                      StructField("sequence", StringType(), nullable)
                      ]
        elif level == 'atom':
            fields = [StructField("structureChainId", StringType(), nullable),
                      StructField("queryLigandId", StringType(), nullable),
                      StructField("queryLigandChainId", StringType(), nullable),
                      StructField("queryLigandNumber", StringType(), nullable),
                      StructField("queryAtomName", StringType(), nullable),
                      StructField("targetChainId", StringType(), nullable),
                      StructField("targetGroupId", StringType(), nullable),
                      StructField("targetGroupNumber", StringType(), nullable),
                      StructField("targetAtomName", StringType(), nullable),
                      StructField("distance", FloatType(), nullable),
                      StructField("sequenceIndex", IntegerType(), nullable),
                      StructField("sequence", StringType(), nullable)
                      ]

        schema = StructType(fields)
        spark = SparkSession.builder.getOrCreate()
        return spark.createDataFrame(row, schema)


class LigandInteractionFingerprint:

    def __init__(self, interaction_filter, level='group'):
        self.filter = interaction_filter
        self.level = level

    def __call__(self, t):
        structure_id = t[0]
        structure = t[1]

        arrays = ColumnarStructure(structure, True)

        # Apply query (ligand) filter
        group_names = arrays.get_group_names()
        qg = self.filter.is_query_group_np(group_names)
        if np.count_nonzero(qg) == 0:
            return []

        elements = arrays.get_elements()
        qe = self.filter.is_query_element_np(elements)
        if np.count_nonzero(qe) == 0:
            return []

        atom_names = arrays.get_atom_names()
        qa = self.filter.is_query_atom_name_np(atom_names)
        if np.count_nonzero(qa) == 0:
            return []

        ### filter prohibited groups??

        # Create mask for polymer atoms
        polymer = arrays.is_polymer()

        # Create mask for ligand atoms
        lig = ~polymer & qg & qe & qa
        if np.count_nonzero(lig) == 0:
            return []

        # Apply target (polymer) filter
        tg = self.filter.is_target_group_np(group_names)
        te = self.filter.is_target_element_np(elements)
        ta = self.filter.is_target_atom_name_np(atom_names)

        poly = polymer & tg & te & ta

        if np.count_nonzero(poly) == 0:
            return []

        chain_names = arrays.get_chain_names()
        group_numbers = arrays.get_group_numbers()
        entity_indices = arrays.get_entity_indices()
        sequence_positions = arrays.get_sequence_positions()

        # Stack coordinates into an nx3 array
        # TODO add this to ColumnarStructure
        c = np.stack((arrays.get_x_coords(), arrays.get_y_coords(), arrays.get_z_coords()), axis=-1)

        # Apply ligand mask to ligand data
        c_ligand = c[lig]
        lg = group_names[lig]
        ln = group_numbers[lig]
        la = atom_names[lig]
        lc = chain_names[lig]

        # Apply polymer mask to polymer data
        c_polymer = c[poly]
        pg = group_names[poly]
        pn = group_numbers[poly]
        pa = atom_names[poly]
        pc = chain_names[poly]
        pt = entity_indices[poly]
        ps = sequence_positions[poly]

        # Calculate distances between polymer and ligand atoms
        poly_tree = cKDTree(c_polymer)
        lig_tree = cKDTree(c_ligand)
        distance_cutoff = self.filter.get_distance_cutoff()
        sparse_dm = poly_tree.sparse_distance_matrix(lig_tree, max_distance=distance_cutoff, output_type='dict')

        # Add interactions to rows.
        # There are redundant interactions when aggregating the results at the 'group' level,
        # since multiple atoms in a group may be involved in interactions.
        # Therefore we use a set of rows to store only unique interactions.
        rows = set([])
        for ind, dis in sparse_dm.items():
            i = ind[0]  # ligand atom index
            j = ind[1]  # polymer atom index
            if self.level == 'group':
                row = Row(structure_id + "." + pc[i],  # structureChainId
                          lg[j],  # queryLigandId
                          lc[j],  # queryLigandChainId
                          ln[j],  # queryLigandNumber
                          pg[i],  # targetGroupId
                          pc[i],  # targetChainId
                          pn[i],  # targetGroupNumber
                          ps[i].item(),  # sequenceIndex
                          structure.entity_list[pt[i]]['sequence']  # sequence
                          )
            elif self.level == 'atom':
                row = Row(structure_id + "." + pc[i],  # structureChainId
                          lg[j],  # queryLigandId
                          lc[j],  # queryLigandChainId
                          ln[j],  # queryLigandNumber
                          la[j],  # queryAtomName
                          pg[i],  # targetGroupId
                          pc[i],  # targetChainId
                          pn[i],  # targetGroupNumber
                          pa[i],  # targetAtomName
                          dis,  # distance
                          ps[i].item(),  # sequenceIndex
                          structure.entity_list[pt[i]]['sequence']  # sequence
                          )

            rows.add(row)

        return rows

#class PolymerInteractionFingerprint(object):
    #  TODO create a KD tree for each chain, then do comparisons?

#class GroupInteractionFingerprint(object):
    # TODO calculate interaction of a group with both polymers and ligands



#!/user/bin/env python
'''containsDnaChain.py

This filter passes entries that contain Dna chains. The default constructor
passes entries that contain at least one Dna chain. If the "exclusive" flag is
set to true in the constructor, all polymer chains must be Dna. For a multi-model
structure (e.g., NMR structure), this filter only checks the first model.

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

from mmtfPyspark.filters import ContainsPolymerChainType


class ContainsDnaChain(object):
    '''Default constructor matches any entry that contains at least one Dna chain.
        As an example, an Dna-protein complex passes this filter.

        Optional constructor that can be used to filter entries that exclusively contain DNA chains.
        For example, with "exclusive" set to true, an Dna-protein complex complex does not pass this filter.

    Attributes
    ----------
    exclusive : bool
       if true, only return entries that contain Dna chains
    '''

    def __init__(self, exclusive=False):
        self.filter = ContainsPolymerChainType(ContainsPolymerChainType.DNA_LINKING, exclusive)

    def __call__(self, t):
        return self.filter(t)



#!/user/bin/env python
__author__ = "Peter Rose"
__version__ = "0.2.0"

from pyspark import SparkFiles
from pyspark.sql import SparkSession
from pyspark.sql.functions import concat, explode, lit, split, upper

# location of dbPTM data
URL = "https://cdn.rcsb.org/resources/protmod/protmod.tsv.gz"
FILE_NAME = "protmod.tsv.gz"


def get_ptm_dataset():
    """
    Retrieves protein modifications present in the Protein Data Bank. Protein Modification
    data are updated weekly in sync with the PDB's weekly update cycle.

    Example:

    >>> ptms = pdbPtmDataset.get_ptm_dataset()
    >>> # filter by residId for: N4-(N-acetylamino)glucosyl-L-asparagine modifications
    >>> # see: http://pir0.georgetown.edu/cgi-bin/resid?id=AA0151
    >>> ptms = ptms.filter("residId = 'AA0151'")
    >>> ptms.show()

    Returns a cached dataset of all post-translational modifications in the PDB. The dataset
    contains the following columns:
    +----------+---------+-------+---------+-------+----+----------+--------------+
    |pdbChainId|pdbResNum|residue|psimodId |residId|ccId|category  |modificationId|
    +----------+---------+-------+---------+-------+----+----------+--------------+
    |1A0H.B    |373      |ASN    |MOD:00831|AA0151 |NAG |attachment|2             |
    |1A0H.B    |584      |NAG    |MOD:00831|AA0151 |NAG |attachment|2             |
    |1A0H.E    |373      |ASN    |MOD:00831|AA0151 |NAG |attachment|2             |
    |1A0H.E    |584      |NAG    |MOD:00831|AA0151 |NAG |attachment|2             |
    |1A14.N    |476A     |NAG    |MOD:00831|AA0151 |NAG |attachment|2             |

    Reference:
    BioJava-ModFinder: identification of protein modifications in 3D structures
    from the Protein Data Bank (2017) Bioinformatics 33: 2047–2049.
    https://doi.org/10.1093/bioinformatics/btx101">doi:10.1093/bioinformatics/btx101

    :return: dataset of PTMs in PDB
    """
    spark = SparkSession.builder.getOrCreate()

    # download dataset
    spark.sparkContext.addFile(URL)

    # read dataset
    ds = spark.read \
        .option('comment', '#') \
        .option('inferSchema', 'true') \
        .option('delimiter', '\t') \
        .csv(SparkFiles.get(FILE_NAME))

    # add column names
    ds = ds.toDF('pdbId', 'chainId', 'modificationId', 'category', 'ccId', 'psimodId', 'residId', 'unused', 'residues')

    # split residue columns into an array
    ds = ds.withColumn('residues', split(ds.residues, ';'))

    # flatten array into individual rows
    ds = ds.withColumn('residues', explode(ds.residues))
    ds = ds.filter("residues != ''")

    # split comma-separated residue information into separate fields
    ds = ds.withColumn('residues', split(ds.residues, ','))
    ds = ds.withColumn('chainId', ds.residues[0])
    ds = ds.withColumn('pdbResNum', ds.residues[1])
    ds = ds.withColumn('residue', ds.residues[3])
    ds = ds.drop(ds.residues)

    # convert PDB Ids to upper case to be consistent with MMTF-Spark convention.
    ds = ds.withColumn('pdbId', upper(ds.pdbId))

    # add a structureChainId column required for joining with other chain-based datasets
    ds = ds.withColumn('pdbChainId', concat(ds.pdbId, lit('.'), ds.chainId))

    return ds.select(ds.pdbChainId, ds.pdbResNum, ds.residue, ds.psimodId, ds.residId, ds.ccId, ds.category,
                     ds.modificationId)



#!/user/bin/env python
'''pdbjMineDataset.py

This filter runs an PDBj Mine 2 Search web service using SQL query

References
----------
- Data are provided through Mine2 SQL: https://pdbj.org/help/mine2-sql
- Queries can be designed with the interactive PDBj Mine 2 query service: https://pdbj.org/mine/sql
- PDB metadata are described in the PDB mmCIF Dictionary: http://mmcif.wwpdb.org/

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

import urllib
import ssl
import tempfile
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, lit, upper, concat
from urllib.request import urlretrieve
import requests


def get_dataset(sqlQuery):
    '''Runs a PDBj Mine 2 search web service using an SQL query

    Parameters
    ----------
    sqlQuery : str
       the sql query for the web service
    '''
    # Create SSl certificate
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE

    # encode SQL
    encodedSQL = urllib.parse.quote(sqlQuery)
    URL = "https://pdbj.org/rest/mine2_sql"

    # Download results to file
    infile = urllib.request.urlopen(URL + "?format=csv&q=" + encodedSQL, context=ctx)
    tmp = tempfile.NamedTemporaryFile(delete=False)
    with open(tmp.name, 'wb') as output:
        output.write(infile.read())

    spark = SparkSession.builder.getOrCreate()

    ds = spark.read.format("csv") \
        .option("header", "true") \
        .option("inferSchema", "true") \
        .option("parserLib", "UNIVOCITY") \
        .load(tmp.name)

    # rename.concatenate columns to assign
    # consistent primary keys to datasets
    if "pdbid" in ds.columns:
        # this project uses upper case pdbids
        ds = ds.withColumn("pdbid", upper(col("pdbid")))

        if "chain" in ds.columns:
            ds = ds.withColumn("structureChainId", \
                               concat(col("pdbid"), lit("."), col("chain")))
            ds.drop("pdbid","chain")
        else:
            ds = ds.withColumnRenamed("pdbid", "structureId")

    return ds



#!/user/bin/env python
'''notFilter.py

This filter wraps another filter and negates its result

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "done"


class NotFilter(object):
    '''Constructor takes another filter as input

    Attributes
    ----------
    filter1 : filter
       first filter to be negated
    filter2 : filter
       second filter to be negated
    '''

    def __init__(self, filter_function):
        self.filter = filter_function

    def __call__(self, t):
        return not self.filter(t)



#!/user/bin/env python
'''resolution.py:

This filter returns true if the resolution value for this
structure is within the sepcified range

References
----------
- `Resolution <http://pdb101.rcsb.org/learn/guide-to-understanding-pdb-data/resolution>`_

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"


class Resolution(object):
    '''This filter returns true if the resolution value of this structure is
    withing the specified range.

    Attributes
    ----------
    min_resolution : float
       The lower bound resolution
    max_resolution : float
       The upper bound resolution

    '''
    def __init__(self, minResolution, maxResolution):
        self.min_Resolution = minResolution
        self.max_Resolution = maxResolution

    def __call__(self, t):
        if t[1].resolution == None:
            return False

        return t[1].resolution >= self.min_Resolution and t[1].resolution <= self.max_Resolution



#!/user/bin/env python
'''chemicalStructureQuery.py

This filter returns entries that contain groups with specified chemical structures (SMILES string).
This chemical structure query supports for query: exact, similar, substructure, and superstructure.
For details see references.

References
----------
- `Chemical Structure Search <http://www.rcsb.org/pdb/staticHelp.do?p=help/advancedsearch/chemSmiles.html>`_

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

from mmtfPyspark.webfilters import AdvancedQuery


class ChemicalStructureQuery(object):

    EXACT = "Exact"
    SIMILAR = "Similar"
    SUBSTRUCTURE = "Substructure"
    SUPERSTRUCTURE = "Superstructure"

    def __init__(self, smiles, queryType="Substructure", percentSimilarity=0.0):
        '''Constructor to setup filter that matches any entry with at least one
        chemical component that matches the specified SMILES string using the
        specified query type.

        For details see:
        `Chemical Structure Search <http://www.rcsb.org/pdb/staticHelp.do?p=help/advancedsearch/chemSmiles.html>`_

        Parameters
        ----------
        smiles : str
           SMILES string representing chemical structure
        queryType : str
           One of the 4 supported types
        percentSimilarity : float
           percent similarity for similarity search. This parameter is ignored
           for all other query types [default: 0.0]
        '''

        if not (queryType == self.EXACT
                or queryType == self.SIMILAR
                or queryType == self.SUBSTRUCTURE
                or queryType == self.SUPERSTRUCTURE):

            raise Exception("Invalid search type: %s" % queryType)

        query = "<orgPdbQuery>" + \
            "<queryType>org.pdb.query.simple.ChemSmilesQuery</queryType>" + \
            "<smiles>" + smiles + "</smiles>" + \
            "<searchType>" + queryType + "</searchType>" + \
            "<similarity>" + str(percentSimilarity) + "</similarity>" + \
            "<polymericType>Any</polymericType>" + \
            "</orgPdbQuery>"

        self.filter = AdvancedQuery(query)

    def __call__(self, t):
        return self.filter(t)



'''proteinSequenceEncoder.py

This class encodes a protein sequence into a feature vector.
The protein sequence must be present in the input data set, the default column
name is "sequence". The default column name for the feature vector is "features".

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

from pyspark.sql import SparkSession
from pyspark.ml.linalg import Vectors
from pyspark.ml.linalg import VectorUDT
from pyspark.ml.feature import Word2Vec, Word2VecModel
from mmtfPyspark.ml import sequenceNgrammer


class ProteinSequenceEncoder(object):
    '''
    This class encodes a protein sequence into a feature vector.
    The protein sequence must be present in the input data set,
    the default column name is "sequence". The default column name
    for the feature vector is "features".

    Attributes
    ----------
    data : DataFrame
       input data to be encoded [None]
    inputCol : str
       name of the input column [sequence]
    outputCol : str
       name of the output column [features]
    '''

    model = None

    AMINO_ACIDS21 = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', \
                     'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'X', 'Y']

    properties = {
        'A' : [1.28,0.05,1.00,0.31,6.11,0.42,0.23],
        'G' : [0.00,0.00,0.00,0.00,6.07,0.13,0.15],
        'V' : [3.67,0.14,3.00,1.22,6.02,0.27,0.49],
        'L' : [2.59,0.19,4.00,1.70,6.04,0.39,0.31],
        'I' : [4.19,0.19,4.00,1.80,6.04,0.30,0.45],
        'F' : [2.94,0.29,5.89,1.79,5.67,0.30,0.38],
        'Y' : [2.94,0.30,6.47,0.96,5.66,0.25,0.41],
        'W' : [3.21,0.41,8.08,2.25,5.94,0.32,0.42],
        'T' : [3.03,0.11,2.60,0.26,5.60,0.21,0.36],
        'S' : [1.31,0.06,1.60,-0.04,5.70,0.20,0.28],
        'R' : [2.34,0.29,6.13,-1.01,10.74,0.36,0.25],
        'K' : [1.89,0.22,4.77,-0.99,9.99,0.32,0.27],
        'H' : [2.99,0.23,4.66,0.13,7.69,0.27,0.30],
        'D' : [1.60,0.11,2.78,-0.77,2.95,0.25,0.20],
        'E' : [1.56,0.15,3.78,-0.64,3.09,0.42,0.21],
        'N' : [1.60,0.13,2.95,-0.60,6.52,0.21,0.22],
        'Q' : [1.56,0.18,3.95,-0.22,5.65,0.36,0.25],
        'M' : [2.35,0.22,4.43,1.23,5.71,0.38,0.32],
        'P' : [2.67,0.00,2.72,0.72,6.80,0.13,0.34],
        'C' : [1.77,0.13,2.43,1.54,6.35,0.17,0.41],
        'X' : [0.00,0.00,0.00,0.00,0.00,0.00,0.00],
    }

	# Source: https://ftp.ncbi.nih.gov/repository/blocks/unix/blosum/BLOSUM/blosum62.blast.new
    blosum62 = {
         #      A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V
        'A' : [ 4,-1,-2,-2, 0,-1,-1, 0,-2,-1,-1,-1,-1,-2,-1, 1, 0,-3,-2, 0],
        'R' : [-1, 5, 0,-2,-3, 1, 0,-2, 0,-3,-2, 2,-1,-3,-2,-1,-1,-3,-2,-3],
        'N' : [-2, 0, 6, 1,-3, 0, 0, 0, 1,-3,-3, 0,-2,-3,-2, 1, 0,-4,-2,-3],
        'D' : [-2,-2, 1, 6,-3, 0, 2,-1,-1,-3,-4,-1,-3,-3,-1, 0,-1,-4,-3,-3],
        'C' : [ 0,-3,-3,-3, 9,-3,-4,-3,-3,-1,-1,-3,-1,-2,-3,-1,-1,-2,-2,-1],
        'Q' : [-1, 1, 0, 0,-3, 5, 2,-2, 0,-3,-2, 1, 0,-3,-1, 0,-1,-2,-1,-2],
        'E' : [-1, 0, 0, 2,-4, 2, 5,-2, 0,-3,-3, 1,-2,-3,-1, 0,-1,-3,-2,-2],
        'G' : [ 0,-2, 0,-1,-3,-2,-2, 6,-2,-4,-4,-2,-3,-3,-2, 0,-2,-2,-3,-3],
        'H' : [-2, 0, 1,-1,-3, 0, 0,-2, 8,-3,-3,-1,-2,-1,-2,-1,-2,-2, 2,-3],
        'I' : [-1,-3,-3,-3,-1,-3,-3,-4,-3, 4, 2,-3, 1, 0,-3,-2,-1,-3,-1, 3],
        'L' : [-1,-2,-3,-4,-1,-2,-3,-4,-3, 2, 4,-2, 2, 0,-3,-2,-1,-2,-1, 1],
        'K' : [-1, 2, 0,-1,-3, 1, 1,-2,-1,-3,-2, 5,-1,-3,-1, 0,-1,-3,-2,-2],
        'M' : [-1,-1,-2,-3,-1, 0,-2,-3,-2, 1, 2,-1, 5, 0,-2,-1,-1,-1,-1, 1],
        'F' : [-2,-3,-3,-3,-2,-3,-3,-3,-1, 0, 0,-3, 0, 6,-4,-2,-2, 1, 3,-1],
        'P' : [-1,-2,-2,-1,-3,-1,-1,-2,-2,-3,-3,-1,-2,-4, 7,-1,-1,-4,-3,-2],
        'S' : [ 1,-1, 1, 0,-1, 0, 0, 0,-1,-2,-2, 0,-1,-2,-1, 4, 1,-3,-2,-2],
        'T' : [ 0,-1, 0,-1,-1,-1,-1,-2,-2,-1,-1,-1,-1,-2,-1, 1, 5,-2,-2, 0],
        'W' : [-3,-3,-4,-4,-2,-2,-3,-2,-2,-3,-2,-3,-1, 1,-4,-3,-2,11, 2,-3],
        'Y' : [-2,-2,-2,-3,-2,-1,-2,-3, 2,-1,-1,-2,-1, 3,-3,-2,-2, 2, 7,-1],
        'V' : [ 0,-3,-3,-3,-1,-2,-2,-3,-3, 3, 1,-2, 1,-1,-2,-2, 0,-3,-1, 4],
        'X' : [-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4,-4],
    }


    def __init__(self, data = None, inputCol = "sequence", outputCol = "features"):

        self.data = data
        self.inputCol = inputCol
        self.outputCol = outputCol


    def one_hot_encode(self, data = None, inputCol = None, outputCol = None):
        '''
        One-hot encodes a protein sequence. The one-hot encoding
        encodes the 20 natural amino acids, plus X for any other
        residue for a total of 21 elements per residue.

        Parameters
        ----------
        data : DataFrame
           input data to be encoded [None]
        inputCol : str
           name of the input column [None]
        outputCol : str
           name of the output column [None]
        '''

        # Setting class variables
        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        session = SparkSession.builder.getOrCreate()
        AMINO_ACIDS21 = self.AMINO_ACIDS21

        # Encoder function to be passed as User Defined Function (UDF)
        def _encoder(s):

            values = [0] * len(AMINO_ACIDS21) * len(s)

            for i in range(len(s)):

                if s[i] in AMINO_ACIDS21:
                    index = AMINO_ACIDS21.index(s[i])

                else:
                    index = AMINO_ACIDS21.index('X')

                values[i*len(AMINO_ACIDS21) + index] = 1

            return Vectors.dense(values)

        session.udf.register("encoder", _encoder, VectorUDT())

        self.data.createOrReplaceTempView("table")
        sql = f"SELECT *, encoder({self.inputCol}) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data


    def property_encode(self, data = None, inputCol = None, outputCol = None):
        '''Encodes a protein sequence by 7 physicochemical properties

        References
        ----------
        Meiler, J., Müller, M., Zeidler, A. et al. J Mol Model (2001)
        https://link.springer.com/article/10.1007/s008940100038

        Parameters
        ----------
        data : DataFrame
           input data to be encoded [None]
        inputCol : str
           name of the input column [None]
        outputCol : str
           name of the output column [None]

        Returns
        -------
        dataset
           dataset with feature vector appended
        '''

        # Setting class variables
        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        session = SparkSession.builder.getOrCreate()
        properties = self.properties

        #Encoder function to be passed as User Defined Function (UDF)
        def _encoder(s):
            values = []

            for i in range(len(s)):

                if s[i] in properties:
                    values += properties[s[i]]

            return Vectors.dense(values)

        session.udf.register("encoder", _encoder, VectorUDT())

        self.data.createOrReplaceTempView("table")
        sql = f"SELECT *, encoder({self.inputCol}) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data


    def blosum62_encode(self, data = None, inputCol = None, outputCol = None):
        '''Encodes a protein sequence by 7 Blosum62

        References
        ----------
        Blosum Matrix
        https://ftp.ncbi.nih.gov/repository/blocks/unix/blosum/BLOSUM/blosum62.blast.new

        Parameters
        ----------
        data : DataFrame
           input data to be encoded [None]
        inputCol : str
           name of the input column [None]
        outputCol : str
           name of the output column [None]

        Returns
        -------
        dataset
           dataset with feature vector appended
        '''

        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        session = SparkSession.builder.getOrCreate()
        blosum62 = self.blosum62

        #Encoder function to be passed as User Defined Function (UDF)
        def _encoder(s):
            values = []

            for i in range(len(s)):

                if s[i] in blosum62:
                    values += blosum62[s[i]]

            return Vectors.dense(values)

        session.udf.register("encoder", _encoder, VectorUDT())

        self.data.createOrReplaceTempView("table")
        sql = f"SELECT *, encoder({self.inputCol}) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data


    def overlapping_ngram_word2vec_encode(self, data = None, inputCol = None,
                                          outputCol = None, n = None,
                                          windowSize = None, vectorSize = None,
                                          fileName = None, sc = None):
        '''Encodes a protein sequence by converting it into n-grams and
        then transforming it into a Word2Vec feature vector.

        If given word2Vec file name, then this function encodes a protein
        sequence by converting it into n-grams and then transforming it using
        pre-trained word2Vec model read from that file

        Parameters
        ----------
        data : DataFrame
           input data to be encoded [None]
        inputCol : str
           name of the input column [None]
        outputCol : str
           name of the output column [None]
        n : int
           The number of words in an n-gram [None]
        windowSize : int
           width of the window used to slide across the  squence, context words from -window to window  [None]
        vectorSize :int
           dimension of the feature vector [None]
        fileName : str
           filename of Word2Vec model [None]

        Returns
        -------
        dataset
           dataset with features vector added to original dataset
        '''

        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        # Create n-grams out of the sequence
        # E.g., 2-gram IDCGH, ... =>[ID, DC, CG, GH, ...]

        data = sequenceNgrammer.ngram(self.data, n, "ngram")

        if not (n == None and windowSize == None and vectorSize == None):
            # Convert n-grams to W2V freature vector
            # [ID, DC, CG, GH, ...] => [0.1234, 0.2394, ...]
            word2Vec = Word2Vec()
            word2Vec.setInputCol("ngram") \
                    .setOutputCol(self.outputCol) \
                    .setNumPartitions(8) \
                    .setWindowSize(windowSize) \
                    .setVectorSize(vectorSize) \

            self.model = word2Vec.fit(data)

        elif fileName != None and sc != None:
            reader = Word2VecModel()

            self.model = reader.load(sc, fileName)

            print(f"model file : {fileName} \n \
                    inputCol : {self.model.getInputCol()} \n \
                    windowSize : {self.model.getWindowSize()} \n \
                    vectorSize : {self.model.getVectorSize()}")

            self.model.setOutputCol(self.outputCol)

        else:
            raise Exception("Either provide word2Vec file (filename) + SparkContext (sc), \
                            or number of words(n) + window size(windowSize) \
                            + vector size (vetorSize), for function parameters")
            return

        return self.model.transform(data)


    def shifted_3gram_word2vec_encode(self, data = None, inputCol = None,
                                      outputCol = None, windowSize = None,
                                      vectorSize = None, fileName = None,
                                      sc = None):
        '''Encodes a protein sequence as three non-overlapping 3-grams,
        trains a Word2Vec model on the 3-grams, and then averages the
        three resulting freature vectors.


        Parameters
        ----------
        data : DataFrame
           input data to be encoded [None]
        inputCol : str
           name of the input column [None]
        outputCol : str
           name of the output column [None]
        windowSize : int
           width of the window used to slide across the sequence context words from -window to window 
        vectorSize : int
           dimension of the feature vector [None]
        fileName : str
           filename of Word2VecModel [None]
        sc : SparkContext
           spark context [None]

        Returns
        -------
        dataset
           dataset with features vector added to original dataset

        References
        ----------
        Asgari E, Mofrad MRK (2015) Continuous Distributed Representation of Biological Sequences for Deep Proteomics and Genomics.  PLOS ONE 10(11): e0141287. doi: https://doi.org/10.1371/journal.pone.0141287
        '''

        if data is not None:
            self.data = data

        if inputCol is not None:
            self.inputCol = inputCol

        if outputCol is not None:
            self.outputCol = outputCol

        if self.data is None:
            raise ValueError("Class variable data is not defined, please pass\
                             in a dataframe into the data parameter")

        # Create n-grams out of the sequence
        # e.g., 2-gram [IDCGH, ...] => [ID. DC, CG, GH,...]

        data = sequenceNgrammer.shifted_ngram(self.data, 3, 0, "ngram0")
        data = sequenceNgrammer.shifted_ngram(data, 3, 1, "ngram1")
        data = sequenceNgrammer.shifted_ngram(data, 3, 2, "ngram2")
        if not (windowSize == None and vectorSize == None):

            ngram0 = data.select("ngram0").withColumnRenamed("ngram0","ngram")
            ngram1 = data.select("ngram1").withColumnRenamed("ngram1","ngram")
            ngram2 = data.select("ngram2").withColumnRenamed("ngram2","ngram")

            ngrams = ngram0.union(ngram1).union(ngram2)

            # Convert n-grams to W2V feature vector
            # [I D, D C, C G, G H, ... ] => [0.1234, 0.2394, .. ]
            word2Vec = Word2Vec()

            word2Vec.setInputCol("ngram") \
                    .setOutputCol("feature") \
                    .setMinCount(10) \
                    .setNumPartitions(8) \
                    .setWindowSize(windowSize) \
                    .setVectorSize(vectorSize)

            self.model = word2Vec.fit(ngrams)

        elif fileName != None and sc != None:
            reader = Word2VecModel()

            self.model = reader.load(sc, fileName)

            print(f"model file : {fileName} \n \
                    inputCol : {self.model.getInputCol()} \n \
                    windowSize : {self.model.getWindowSize()} \n \
                    vectorSize : {self.model.getVectorSize()}")

        else:
            raise Exception("Either provide word2Vec file (filename) + SparkContext (sc), \
                            or window size(windowSize) + vector size (vetorSize), \
                            for function parameters")
            return

        #data = data.withColumn("feature0",self.model.transform(data.select('ngram0').withColumnRenamed("ngram0","ngram")))
        for i in reversed(range(3)):
            feature = self.model.transform(data.select('ngram' + str(i)).withColumnRenamed("ngram" + str(i),"ngram"))
            data = data.join(feature.withColumnRenamed("ngram","ngram" + str(i)), "ngram" + str(i))
            data = data.withColumnRenamed("feature", "feature" + str(i))


        data = self._average_feature_vectors(data, self.outputCol)
        data.printSchema()

        cols = ['structureChainId','sequence','labelQ8','labelQ3','ngram0','ngram1',\
                'ngram2','feature0','feature1','feature2', 'features']

        data = data.select(cols)


        return data


    def get_word2vec_model(self):
        '''Returns a Word2VecModel created by overlapping_ngram_word2vec_encode()

        Returns
        -------
        model
           overlapping Ngram Word2VecModel if available, otherwise None
        '''

        return self.model


    def _average_feature_vectors(self, data, outputCol):
        '''Average the feature vectors

        Parameters
        ----------
        data : DataFrame
           input dataframe
        outputCol : str
           name of the output column
        '''

        session = SparkSession.builder.getOrCreate()

        def _averager(v1, v2, v3):
            f1 = v1.toArray()
            f2 = v2.toArray()
            f3 = v3.toArray()

            length = min(len(f1), len(f2), len(f3))
            average = []

            for i in range(length):
                average.append((f1[i] + f2[i] + f3[i])/3.0)

            return Vectors.dense(average)

        session.udf.register("averager", _averager, VectorUDT())

        data.createOrReplaceTempView("table")

        sql = f"SELECT *, averager(feature0, feature1, feature2) AS {self.outputCol} from table"

        data = session.sql(sql)

        return data



#!/user/bin/env python
__author__ = "Peter Rose"
__version__ = "0.2.0"

from pyspark.sql import DataFrame
from pyspark.sql.functions import explode
from pyspark.sql.types import ArrayType


# /**
#  * Removes spaces from column names to ensure compatibility with parquet
#  * files.
#  *
#  * @param original
#  *            dataset
#  * @return dataset with columns renamed
#  */
#


def remove_spaces_from_column_names(dataset: DataFrame):
    for name in dataset.columns:
        new_name = name.replace(" ", "")
        original = dataset.withColumnRenamed(name, new_name)

        return original


def flatten_dataset(dataset: DataFrame):
    tmp = dataset
    for field in tmp.schema.fields:
        if isinstance(field.dataType, ArrayType):
            print(field.name, field.dataType)
            tmp = tmp.withColumn(field.name, explode(tmp.field.name))

    return tmp



#!/user/bin/env python
'''drugBankDatset.py

This module provides access to DrugBank containing drug structure and drug target
imformation. These datasets contain identifiers and names for integration with
other data resources.

References
----------
- Drug Bank. https://www.drugbank.ca
- Wishart DS, et al., DrugBank 5.0: a major update to the DrugBank database for 2018.
  Nucleic Acids Res. 2017 Nov 8. https://dx.doi.org/10.1093/nar/gkx1037

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

import requests
import tempfile
import io
from zipfile import ZipFile
from io import BytesIO
from pyspark.sql import SparkSession

DRUG_GROUP = ['ALL', 'APPROVED', 'EXPERIMENTAL', 'NUTRACEUTICAL', 'ILLICIT'\
              'WITHDRAWN', 'INVESTIGATIONAL']
DRUG_TYPE = ['SMALL_MOLECULE', 'BIOTECH']
BASE_URL = "https://www.drugbank.ca/releases/latest/downloads/"
BUFFER = 2048


def get_open_drug_links():
    '''Downloads the DrugBank Open Data dataset with drug structure external
    links and identifiers. See DrugBank.

    This dataset contains drug common names, synonyms, CAS numbers, and
    Standard InChIKeys.

    The DrugBank Open Data dataset is a public domain dataset that can be
    used freely in your application or project (including commercial use).
    It is released under a Creative Common’s CC0 International License.

    References
    ----------
    Open Data dataset. https://www.drugbank.ca/releases/latest#open-data

    Examples
    --------
    Get DrugBank open dataset:

    >>> openDrugLinks = DrugBankDataset.get_open_drug_links()
    >>> openDrugLinks.show()
    +----------+--------------------+-----------+--------------------+
    |DrugBankID|          Commonname|        CAS|    StandardInChIKey|
    +----------+--------------------+-----------+--------------------+
    |   DB00006|         Bivalirudin|128270-60-0|OIRCOABEOLEUMC-GE...|
    |   DB00014|           Goserelin| 65807-02-5|BLCLNMBMMGCOAS-UR...|
    +----------+--------------------+-----------+--------------------+

    Returns
    -------
    dataset
       DrugBank Dataset

    '''

    url = BASE_URL + "all-drugbank-vocabulary"
    return get_dataset(url)


def get_drug_links(drugGroup, username, password):
    '''Downloads drug structure external links and identifiers from DrugBank.
    Either all or subsets of data can be downloaded by specifying the
    DrugGroup:
        ALL, APPROVED, EXPERIMENTAL, NUTRACEUTICAL, ILLLICT, WITHDRAWN,
        INVESTIGATIONAL.

    The structure external links datasets include drug structure information
    in the form of InChI/InChI Key/SMILES as well as identifiers for other
    drug-structure resources (such as ChEBI, ChEMBL,ChemSpider, BindingDB,
    etc.). Included in each dataset is also the PubChem Compound ID (CID) and
    the particular PubChem Substance ID (SID) for the given DrugBank record.

    These DrugBank datasets are released under a Creative Common’s
    Attribution-NonCommercial 4.0 International License. They can be used
    freely in your non-commercial application or project. A DrugBank user
    account and authentication is required to download these datasets.

    References
    ----------
    External Drug Links:
        https://www.drugbank.ca/releases/latest#external-links

    Examples
    --------
    Get dataset of external links and identifiers of approved drugs:

    >>> username = "<your DrugBank username>"
    >>> String password = "<your DrugBank password>"
    >>> drugLinks = get_drug_links("APPROVED", username, password)
    >>> drugLinks.show()

    Parameters
    ----------
    durgGroup : str
       specific dataset to be downloaded, has to be in the pre-defined DURG_GROUP list
    usesrname : str
       DrugBank username
    password : str
       DrugBank password

    Returns
    -------
    dataset
       DrugBank Dataset
    '''

    if drugGroup.upper() not in DRUG_GROUP:
        raise ValueError("drugGroup not in pre-defined durgGroups")

    url = BASE_URL + drugGroup + "-structure-links"
    return get_dataset(url, username, password)


def get_drug_target_links(drug, username, password):
    '''Downloads drug target external links and identifiers from DrugBank.
    Either all or subsets of data can be downloaded by specifying the
    DrugGroup:
        ALL, APPROVED, EXPERIMENTAL, NUTRACEUTICAL, ILLLICT, WITHDRAWN,
        INVESTIGATIONAL.
    OR DrugType:
        SMALL_MOLECULE, BIOTECH.

    The drug target external links datasets include drug name, drug type
    (small molecule, biotech), UniProtID and UniProtName.

    These DrugBank datasets are released under the Creative Common’s
    Attribution-NonCommercial 4.0 International License. They can be used
    freely in your non-commercial application or project. A DrugBank user
    account and authentication is required to download these datasets.

    References
    ----------
    Target Drug-UniProt:
        https://www.drugbank.ca/releases/latest#external-links

    Examples
    --------
    Get dataset of drug target external links and identifiers of all drugs in DrugBank:

    >>> username = "<your DrugBank username>"
    >>> password = "<your DrugBank password>"
    >>> drugTargetLinks = get_drug_target_links("ALL",
    ...                                         username,
    ...                                         password)
    >>> drugTargetLinks.show()

    Parameters
    ----------
    durg : str
       specific dataset to be downloaded, has to be either in
       the DrugGroup list OR DrugType list.
    usesrname : str
       DrugBank username
    password : str
       DrugBank password

    Returns
    -------
    dataset
       DrugBank Dataset
    '''

    if drug.upper() in DRUG_GROUP or drug.upper() in DRUG_TYPE:
        url = BASE_URL + "target-" + drug + "-uniprot-links"
    else:
        raise ValueError("drug not in pre-defined durgGroups or drugTypes")

    return get_dataset(url, username, password)


def get_dataset(url, username=None, password=None):
    '''Downloads a DrugBank dataset

    Parameters
    ----------
    url : str
       DrugBank dataset download links
    username : str, optional
       DrugBank username <None>
    password : str, optional
       DrugBank password <None>

    Returns
    -------
    dataset
       DrugBank dataset
    '''
    if username is None and password is None:
        # get input stream to first zip entry
        req = requests.get(url)
    else:
        # TODO dataset that requires authentication
        req = requests.get(url, auth=(username, password))
        if req.text ==  'Invalid Email or password.':
                raise ValueError('Invalid Email or password.')

    # Decode and unzip file
    unzipped = _decode_as_zip_input_stream(req.content)

    # save data to a temporary file (Dataset csv reader requires a input
    # file!)
    tempFileName = _save_temp_file(unzipped)

    #load temporary CSV file to Spark dataste
    dataset = _read_csv(tempFileName)
    dataset = _remove_spaces_from_column_names(dataset)

    return dataset


def _decode_as_zip_input_stream(content):
    '''Returns an input stream to the first zip file entry

    Parameters
    ----------
    content : inputStream
       inputStream content from request

    Returns
    -------
    inputStream
       unzipped InputStream
    '''

    zipfile = ZipFile(BytesIO(content))
    return [line.decode() for line \
            in zipfile.open(zipfile.namelist()[0]).readlines()]


def _save_temp_file(unzipped):
    '''Saves tabular report as a temporary CSV file

    Parameters
    ----------
    unzipped : list
       list of unzipped content

    Returns
    -------
    str
       path to the tempfile
    '''
    tempFile = tempfile.NamedTemporaryFile(delete=False)
    with io.open(tempFile.name, "w", encoding='utf-8') as t:
       t.writelines(unzipped)

    return tempFile.name


def _read_csv(inputFileName):
    '''Reads CSV file into Spark dataset

    Parameters
    ----------
    fileName : str
       name of the input csv fileName

    Returns
    -------
    dataset
       a spark dataset
    '''
    spark = SparkSession.builder.getOrCreate()
    dataset = spark.read.format("csv") \
                        .option("header", "true") \
                        .option("inferSchema", "true") \
                        .load(inputFileName)
    return dataset


def _remove_spaces_from_column_names(original):
    '''Remove spaces from column names to ensure compatibility with parquet

    Parameters
    ----------
    original : dataset
       the original dataset

    Returns
    -------
    dataset
       dataset with columns renamed
    '''

    for existingName in original.columns:
        newName = existingName.replace(' ','')
        # TODO: double check dataset "withColumnRenamed" funciton
        original = original.withColumnRenamed(existingName,newName)

    return original



#!/user/bin/env python
'''pisces.py

This filter passes through representative structures or protein chains
from the PISCES CulledPDB sets. A CulledPDB set is selected by specifying
sequenceIdentity and resolution cutoff values from the following
list:
- sequenceIdentity = [20, 25, 30, 40, 50, 60, 70, 80, 90]
- resolution = [1.6, 1.8, 2.0, 2.2, 2.5, 3.0]

References
----------
- `PISCES <http://dunbrack.fccc.edu/PISCES.php>`_
- G. Wang and R. L. Dunbrack, Jr. PISCES: a protein sequence culling server.  Bioinformatics, 19:1589-1591, 2003.

'''
__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "Done"

from mmtfPyspark.webservices import PiscesDownloader


class Pisces(object):
    '''Filters representative PDB structures and polymer chains based
    on the specified criteria using PISCES CulledPDB sets.

    sequenceIdentity = 20, 25, 30, 40, 50, 60, 70, 80, 90
    resolution = 1.6, 1.8, 2.0, 2.2, 2.5, 3.0

    Attributes
    ----------
    sequenceIdentity : int
       sequence identity cutoff values
    resolution : float
       resolution cutoff value
    '''

    def __init__(self, sequenceIdentity, resolution):
        self.pdbIds = set()

        pD = PiscesDownloader(sequenceIdentity, resolution)

        for pdbId in pD.get_structure_chain_ids():
            self.pdbIds.add(pdbId)
            self.pdbIds.add(pdbId[:4])

    def __call__(self, t):
        return t[0] in self.pdbIds



#!/user/bin/env python
'''sequenceSimilarity.py

This filter returns entries that pass the sequence similarity search
criteria. Searches protein and nucleic acid sequences using the BLAST.
PSI-BLAST is used to find more distantly related protein sequences.

The E value, or Expect value, is a parameter that describes the number of
hits one can expect to see just by chance when searching a database of a
particular size. For example, an E value of one indicates that a result will
contain one sequence with similar score simply by chance. The scoring takes
chain length into consideration and therefore shorter sequences can have
identical matches with high E value.

The Low Complexity filter masks low complexity regions in a sequence to
filter out avoid spurious alignments.

Sequence Identity Cutoff (%) filter removes entries of low sequence
similarity. The cutoff value is a percentage value between 0 to 100.

Note: sequences must be at least 12 residues long. For shorter sequences try
the Sequence Motif Search.

References
----------
- BLAST: BLAST: Sequence searching using NCBI's BLAST (Basic Local Alignment
  Search Tool) Program , Altschul, S.F., Gish, W., Miller, W., Myers, E.W.
  and Lipman, D.J. Basic local alignment search tool. J. Mol. Biol. 215:
  403-410 (1990)

- PSI-BLAST: Sequence searching to detect distantly related evolutionary
  relationships using NCBI's PSI-BLAST (Position-Specific Iterated BLAST).
'''

__author__ = "Mars (Shih-Cheng) Huang"
__maintainer__ = "Mars (Shih-Cheng) Huang"
__email__ = "marshuang80@gmail.com"
__version__ = "0.2.0"
__status__ = "done"

from mmtfPyspark.webfilters import AdvancedQuery


class SequenceSimilarity(object):

    BLAST = 'blast'
    PSI_BLAST = 'psi-blast'

    def __init__(self, sequence, searchTool=BLAST, eValueCutoff=10.0,
                 sequenceIdentityCutoff=0, maskLowComplexity=True):
        '''Filters by squence similarity using all default parameters.

        Parameters
        ----------
        sequence : str
           query sequence
        searchTool : class variable
           sequenceSimilarity.BLAST or sequenceSimilarity.PSI_BLAST
        eValueCutoff : float
           maximun e-value
        sequenceIdentityCutoff : int
           minimum sequence identity cutoff
        maskLowComplexity : bool
           if true, mask (ignore) low complexity regions in sequence
        '''

        if len(sequence) < 12:
            raise ValueError(
                "ERROR: the query sequence must be at least 12 residues long")

        complexity = 'yes' if maskLowComplexity else 'no'

        query = "<orgPdbQuery>" + "<queryType>org.pdb.query.simple.SequenceQuery</queryType>" \
                + "<sequence>" + sequence + "</sequence>" + "<searchTool>" + searchTool \
                + "</searchTool>" + "<maskLowComplexity>" + complexity\
                + "</maskLowComplexity>" + "<eValueCutoff>" + str(eValueCutoff) \
                + "</eValueCutoff>" + "<sequenceIdentityCutoff>" + str(sequenceIdentityCutoff) \
                + "</sequenceIdentityCutoff>" + "</orgPdbQuery>"

        self.filter = AdvancedQuery(query)

    def __call__(self, t):
        return self.filter(t)
